@@ -8,16 +8,17 @@ from sql.conditionals import Coalesce
 from sql.functions import Substring
 from sql.operators import Concat, Equal
 
+from trytond.cache import Cache
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, MatchMixin, DeactivableMixin, fields,
-    sequence_ordered, Exclude)
+    DeactivableMixin, Exclude, MatchMixin, ModelSQL, ModelView, fields,
+    sequence_ordered)
 from trytond.model.exceptions import AccessError
-from trytond.pyson import Eval, If
 from trytond.pool import Pool
+from trytond.pyson import Eval, If
 from trytond.rpc import RPC
 from trytond.transaction import Transaction
-from trytond.cache import Cache
+
 from .exceptions import InvalidFormat
 
 
@@ -27,14 +28,17 @@ class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
     party = fields.Many2One('party.party', 'Party', required=True,
         ondelete='CASCADE', select=True, states={
             'readonly': Eval('id', 0) > 0,
-            },
-        depends=['id'])
+            })
     party_name = fields.Char(
         "Party Name",
         help="If filled, replace the name of the party for address formatting")
     name = fields.Char("Building Name")
     street = fields.Text("Street")
-    zip = fields.Char("Zip")
+    street_single_line = fields.Function(
+        fields.Char("Street"),
+        'on_change_with_street_single_line',
+        searcher='search_street_single_line')
+    postal_code = fields.Char("Postal Code")
     city = fields.Char("City")
     country = fields.Many2One('country.country', "Country")
     subdivision_types = fields.Function(
@@ -49,17 +53,22 @@ class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
                 ('type', 'in', Eval('subdivision_types', [])),
                 ()
                 ),
-            ],
-        depends=['country', 'subdivision_types'])
+            ])
     full_address = fields.Function(fields.Text('Full Address'),
             'get_full_address')
+    identifiers = fields.One2Many(
+        'party.identifier', 'address', "Identifiers",
+        domain=[
+            ('party', '=', Eval('party')),
+            ('type', 'in', ['fr_siret']),
+            ])
 
     @classmethod
     def __setup__(cls):
         super(Address, cls).__setup__()
         cls._order.insert(0, ('party', 'ASC'))
         cls.__rpc__.update(
-            autocomplete_zip=RPC(instantiate=0, cache=dict(days=1)),
+            autocomplete_postal_code=RPC(instantiate=0, cache=dict(days=1)),
             autocomplete_city=RPC(instantiate=0, cache=dict(days=1)),
             )
 
@@ -67,6 +76,10 @@ class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
     def __register__(cls, module_name):
         cursor = Transaction().connection.cursor()
         sql_table = cls.__table__()
+        table = cls.__table_handler__(module_name)
+
+        # Migration from 5.8: rename zip to postal code
+        table.column_rename('zip', 'postal_code')
 
         super(Address, cls).__register__(module_name)
 
@@ -82,6 +95,15 @@ class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
                     [value]))
             table.drop_column('streetbis')
 
+    @fields.depends('street')
+    def on_change_with_street_single_line(self, name=None):
+        if self.street:
+            return " ".join(self.street.splitlines())
+
+    @classmethod
+    def search_street_single_line(cls, name, domain):
+        return [('street',) + tuple(domain[1:])]
+
     _autocomplete_limit = 100
 
     @fields.depends('country', 'subdivision')
@@ -99,25 +121,25 @@ class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
 
     def _autocomplete_search(self, domain, name):
         pool = Pool()
-        Zip = pool.get('country.zip')
+        PostalCode = pool.get('country.postal_code')
         if domain:
-            records = Zip.search(domain, limit=self._autocomplete_limit)
+            records = PostalCode.search(domain, limit=self._autocomplete_limit)
             if len(records) < self._autocomplete_limit:
                 return sorted({getattr(z, name) for z in records})
         return []
 
     @fields.depends('city', methods=['_autocomplete_domain'])
-    def autocomplete_zip(self):
+    def autocomplete_postal_code(self):
         domain = self._autocomplete_domain()
         if self.city:
             domain.append(('city', 'ilike', '%%%s%%' % self.city))
-        return self._autocomplete_search(domain, 'zip')
+        return self._autocomplete_search(domain, 'postal_code')
 
-    @fields.depends('zip', methods=['_autocomplete_domain'])
+    @fields.depends('postal_code', methods=['_autocomplete_domain'])
     def autocomplete_city(self):
         domain = self._autocomplete_domain()
-        if self.zip:
-            domain.append(('zip', 'ilike', '%s%%' % self.zip))
+        if self.postal_code:
+            domain.append(('postal_code', 'ilike', '%s%%' % self.postal_code))
         return self._autocomplete_search(domain, 'city')
 
     def get_full_address(self, name):
@@ -129,27 +151,35 @@ class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
             filter(None, (x.strip() for x in full_address.splitlines())))
 
     def _get_address_substitutions(self):
+        pool = Pool()
+        Country = pool.get('country.country')
+
         context = Transaction().context
         subdivision_code = ''
         if getattr(self, 'subdivision', None):
             subdivision_code = self.subdivision.code or ''
             if '-' in subdivision_code:
                 subdivision_code = subdivision_code.split('-', 1)[1]
+        country_name = ''
+        if getattr(self, 'country', None):
+            with Transaction().set_context(language='en'):
+                country_name = Country(self.country.id).name
         substitutions = {
             'party_name': '',
             'attn': '',
             'name': getattr(self, 'name', None) or '',
             'street': getattr(self, 'street', None) or '',
-            'zip': getattr(self, 'zip', None) or '',
+            'postal_code': getattr(self, 'postal_code', None) or '',
             'city': getattr(self, 'city', None) or '',
             'subdivision': (self.subdivision.name
                 if getattr(self, 'subdivision', None) else ''),
             'subdivision_code': subdivision_code,
-            'country': (self.country.name
-                if getattr(self, 'country', None) else ''),
+            'country': country_name,
             'country_code': (self.country.code or ''
                 if getattr(self, 'country', None) else ''),
             }
+        # Keep zip for backward compatibility
+        substitutions['zip'] = substitutions['postal_code']
         if context.get('address_from_country') == getattr(self, 'country', ''):
             substitutions['country'] = ''
         if context.get('address_with_party', False):
@@ -172,8 +202,8 @@ class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
 
     def get_rec_name(self, name):
         party = self.party_full_name
-        if self.street:
-            street = self.street.splitlines()[0]
+        if self.street_single_line:
+            street = self.street_single_line
         else:
             street = None
         if self.country:
@@ -185,7 +215,7 @@ class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
                     party,
                     self.name,
                     street,
-                    self.zip,
+                    self.postal_code,
                     self.city,
                     country]))
 
@@ -199,7 +229,7 @@ class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
             ('party',) + tuple(clause[1:]),
             ('name',) + tuple(clause[1:]),
             ('street',) + tuple(clause[1:]),
-            ('zip',) + tuple(clause[1:]),
+            ('postal_code',) + tuple(clause[1:]),
             ('city',) + tuple(clause[1:]),
             ('country',) + tuple(clause[1:]),
             ]
@@ -246,7 +276,7 @@ class AddressFormat(DeactivableMixin, MatchMixin, ModelSQL, ModelView):
         "- ${name}\n"
         "- ${attn}\n"
         "- ${street}\n"
-        "- ${zip}\n"
+        "- ${postal_code}\n"
         "- ${city}\n"
         "- ${subdivision}\n"
         "- ${subdivision_code}\n"
@@ -297,10 +327,11 @@ class AddressFormat(DeactivableMixin, MatchMixin, ModelSQL, ModelView):
 
     @classmethod
     def default_format_(cls):
-        return """${party_name}
+        return """${attn}
+${party_name}
 ${name}
 ${street}
-${zip} ${city}
+${postal_code} ${city}
 ${subdivision}
 ${COUNTRY}"""
 
@@ -321,22 +352,25 @@ ${COUNTRY}"""
         cls._get_format_cache.clear()
 
     @classmethod
-    def validate(cls, formats):
-        super(AddressFormat, cls).validate(formats)
-        for format_ in formats:
-            format_.check_format()
+    def validate_fields(cls, formats, field_names):
+        super().validate_fields(formats, field_names)
+        cls.check_format(formats, field_names)
 
-    def check_format(self):
+    @classmethod
+    def check_format(cls, formats, field_names=None):
         pool = Pool()
         Address = pool.get('party.address')
+        if field_names and 'format_' not in field_names:
+            return
         address = Address()
-        try:
-            Template(self.format_).substitute(
-                **address._get_address_substitutions())
-        except Exception as exception:
-            raise InvalidFormat(gettext('party.invalid_format',
-                    format=self.format_,
-                    exception=exception)) from exception
+        substitutions = address._get_address_substitutions()
+        for format_ in formats:
+            try:
+                Template(format_.format_).substitute(**substitutions)
+            except Exception as exception:
+                raise InvalidFormat(gettext('party.invalid_format',
+                        format=format_.format_,
+                        exception=exception)) from exception
 
     @classmethod
     def get_format(cls, address, pattern=None):
