@@ -1,28 +1,37 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from stdnum import get_cc_module
 import stdnum.exceptions
-from sql import Null, Column, Literal
-from sql.functions import CharLength, Substring, Position
+from sql import Column, Literal, Null
+from sql.aggregate import Min
+from sql.functions import CharLength, Position, Substring
+from stdnum import get_cc_module
 
-from trytond.i18n import gettext
-from trytond.model import (ModelView, ModelSQL, MultiValueMixin, ValueMixin,
-    DeactivableMixin, fields, Unique, sequence_ordered)
-from trytond.wizard import Wizard, StateTransition, StateView, Button
-from trytond.pyson import Eval, Bool
-from trytond.transaction import Transaction
-from trytond.pool import Pool
 from trytond import backend
-from trytond.tools.multivalue import migrate_property
+from trytond.i18n import gettext
+from trytond.model import (
+    DeactivableMixin, ModelSQL, ModelView, MultiValueMixin, Unique, ValueMixin,
+    fields, sequence_ordered)
+from trytond.model.exceptions import AccessError
+from trytond.pool import Pool
+from trytond.pyson import Bool, Eval
 from trytond.tools import lstrip_wildcard
+from trytond.tools.multivalue import migrate_property
+from trytond.transaction import Transaction
+from trytond.wizard import Button, StateTransition, StateView, Wizard
+
+from .contact_mechanism import _PHONE_TYPES
 from .exceptions import (
-    InvalidIdentifierCode, VIESUnavailable, SimilarityWarning, EraseError)
+    EraseError, InvalidIdentifierCode, SimilarityWarning, VIESUnavailable)
 
 
 class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
     "Party"
     __name__ = 'party.party'
 
+    _contact_mechanism_states = {
+        'readonly': Eval('id', -1) >= 0,
+        }
+
     name = fields.Char(
         "Name", select=True,
         help="The main identifier of the party.")
@@ -30,7 +39,6 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
         states={
             'readonly': Eval('code_readonly', True),
             },
-        depends=['code_readonly'],
         help="The unique identifier of the party.")
     code_readonly = fields.Function(fields.Boolean('Code Readonly'),
         'get_code_readonly')
@@ -58,11 +66,24 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
             },
         help="The party replacing this one.")
     full_name = fields.Function(fields.Char('Full Name'), 'get_full_name')
-    phone = fields.Function(fields.Char('Phone'), 'get_mechanism')
-    mobile = fields.Function(fields.Char('Mobile'), 'get_mechanism')
-    fax = fields.Function(fields.Char('Fax'), 'get_mechanism')
-    email = fields.Function(fields.Char('E-Mail'), 'get_mechanism')
-    website = fields.Function(fields.Char('Website'), 'get_mechanism')
+    phone = fields.Function(
+        fields.Char("Phone", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
+    mobile = fields.Function(
+        fields.Char("Mobile", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
+    fax = fields.Function(
+        fields.Char("Fax", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
+    email = fields.Function(
+        fields.Char("E-Mail", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
+    website = fields.Function(
+        fields.Char("Website", states=_contact_mechanism_states),
+        'get_contact_mechanism', setter='set_contact_mechanism')
+    distance = fields.Function(fields.Integer('Distance'), 'get_distance')
+
+    del _contact_mechanism_states
 
     @classmethod
     def __setup__(cls):
@@ -71,11 +92,11 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
         cls._sql_constraints = [
             ('code_uniq', Unique(t, t.code), 'party.msg_party_code_unique')
             ]
-        cls._order.insert(0, ('name', 'ASC'))
+        cls._order.insert(0, ('distance', 'ASC NULLS LAST'))
+        cls._order.insert(1, ('name', 'ASC'))
         cls.active.states.update({
                 'readonly': Bool(Eval('replaced_by')),
                 })
-        cls.active.depends.append('replaced_by')
 
     @classmethod
     def __register__(cls, module_name):
@@ -119,17 +140,7 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
 
     @classmethod
     def tax_identifier_types(cls):
-        return [
-            'ad_nrt', 'al_nipt', 'ar_cuit', 'be_vat', 'bg_vat', 'ch_vat',
-            'cl_rut', 'co_rut', 'cz_dic', 'de_vat', 'do_rnc',
-            'dk_cvr', 'ec_ruc', 'ee_kmkr', 'es_cif', 'es_nie', 'es_nif',
-            'eu_vat', 'fi_alv', 'fr_tva', 'gb_vat', 'gr_vat', 'gt_nit',
-            'hu_anum', 'ie_vat', 'is_vsk', 'it_iva', 'jp_cn', 'lt_pvm',
-            'lu_tva', 'lv_pvn', 'mc_tva', 'md_idno', 'mt_vat', 'mx_rfc',
-            'nl_btw', 'no_mva', 'nz_ird', 'pe_ruc', 'pl_nip', 'pt_nif',
-            'py_ruc', 'ro_cf', 'rs_pib', 'ru_inn', 'se_vat', 'si_ddv',
-            'sk_dph', 'sm_coe', 'us_atin', 'us_ein', 'us_itin', 'us_ptin',
-            'us_ssn', 'us_tin', 'uy_ruc', 've_rif', 'za_tin']
+        return TAX_IDENTIFIER_TYPES
 
     def get_tax_identifier(self, name):
         types = self.tax_identifier_types()
@@ -162,21 +173,96 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
     def get_full_name(self, name):
         return self.name
 
-    def get_mechanism(self, name):
+    def get_contact_mechanism(self, name):
         for mechanism in self.contact_mechanisms:
             if mechanism.type == name:
                 return mechanism.value
         return ''
 
+    @classmethod
+    def set_contact_mechanism(cls, parties, name, value):
+        pool = Pool()
+        ContactMechanism = pool.get('party.contact_mechanism')
+        contact_mechanisms = []
+        for party in parties:
+            if getattr(party, name):
+                type_string = cls.fields_get([name])[name]['string']
+                raise AccessError(gettext(
+                        'party.msg_party_set_contact_mechanism',
+                        party=party.rec_name,
+                        field=type_string))
+            if value:
+                contact_mechanisms.append(ContactMechanism(
+                        party=party,
+                        type=name,
+                        value=value))
+        ContactMechanism.save(contact_mechanisms)
+
+    @classmethod
+    def _distance_query(cls, usages=None, party=None, depth=None):
+        context = Transaction().context
+        if party is None:
+            party = context.get('related_party')
+
+        if not party:
+            return
+
+        table = cls.__table__()
+        return table.select(
+            table.id.as_('to'),
+            Literal(0).as_('distance'),
+            where=(table.id == party))
+
+    @classmethod
+    def get_distance(cls, parties, name):
+        distances = {p.id: None for p in parties}
+        query = cls._distance_query()
+        if query:
+            cursor = Transaction().connection.cursor()
+            cursor.execute(*query.select(
+                    query.to.as_('to'),
+                    Min(query.distance).as_('distance'),
+                    group_by=[query.to]))
+            distances.update(cursor)
+        return distances
+
+    @classmethod
+    def order_distance(cls, tables):
+        party, _ = tables[None]
+        key = 'distance'
+        if key not in tables:
+            query = cls._distance_query()
+            if not query:
+                return []
+            query = query.select(
+                    query.to.as_('to'),
+                    Min(query.distance).as_('distance'),
+                    group_by=[query.to])
+            join = party.join(query, type_='LEFT',
+                condition=query.to == party.id)
+            tables[key] = {
+                None: (join.right, join.condition),
+                }
+        else:
+            query, _ = tables[key][None]
+        return [query.distance]
+
+    @classmethod
+    def index_set_field(cls, name):
+        index = super().index_set_field(name)
+        if name in _PHONE_TYPES:
+            # Phone validation may need the address country
+            index = cls.index_set_field('addresses') + 1
+        return index
+
     @classmethod
     def _new_code(cls, **pattern):
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Configuration = pool.get('party.configuration')
         config = Configuration(1)
         sequence = config.get_multivalue('party_sequence', **pattern)
         if sequence:
-            return Sequence.get_id(sequence.id)
+            return sequence.get()
 
     @classmethod
     def create(cls, vlist):
@@ -307,138 +393,243 @@ class PartyCategory(ModelSQL):
         ondelete='CASCADE', required=True, select=True)
 
 
-class Identifier(sequence_ordered(), ModelSQL, ModelView):
+IDENTIFIER_TYPES = [
+    ('ad_nrt', "Andorra Tax Number"),
+    ('al_nipt', "Albanian VAT Number"),
+    ('ar_cuit', "Argentinian Tax Number"),
+    ('ar_dni', "Argentinian National Identity Number"),
+    ('at_businessid', "Austrian Company Register"),
+    ('at_tin', "Austrian Tax Identification"),
+    ('at_uid', "Austrian Umsatzsteuer-Identifikationsnummer"),
+    ('at_vnr', "Austrian Social Security Number"),
+    ('au_abn', "Australian Business Number"),
+    ('au_acn', "Australian Company Number"),
+    ('au_tfn', "Australian Tax File Number"),
+    ('be_vat', "Belgian Enterprise Number"),
+    ('bg_egn', "Bulgarian Personal Identity Codes"),
+    ('bg_pnf', "Bulgarian Number of a Foreigner"),
+    ('bg_vat', "Bulgarian VAT Number"),
+    ('br_cnpj', "Brazillian Company Identifier"),
+    ('br_cpf', "Brazillian National Identifier"),
+    ('by_unp', "Belarus VAT Number"),
+    ('ca_bn', "Canadian Business Number"),
+    ('ca_sin', "Canadian Social Insurance Number"),
+    ('ch_ssn', "Swiss Social Security Number"),
+    ('ch_uid', "Swiss Business Identifier"),
+    ('ch_vat', "Swiss VAT Number"),
+    ('cl_rut', "Chilean National Tax Number"),
+    ('cn_ric', "Chinese Resident Identity Card Number"),
+    ('cn_uscc', "Chinese Unified Social Credit Code"),
+    ('co_nit', "Colombian Identity Code"),
+    ('co_rut', "Colombian Business Tax Number"),
+    ('cr_cpf', "Costa Rica Physical Person ID Number"),
+    ('cr_cpj', "Costa Rica Tax Number"),
+    ('cr_cr', "Costa Rica Foreigners ID Number"),
+    ('cu_ni', "Cuban Identity Card Number"),
+    ('cy_vat', "Cypriot VAT Number"),
+    ('cz_dic', "Czech VAT Number"),
+    ('cz_rc', "Czech National Identifier"),
+    ('de_handelsregisternummer', "German Company Register Number"),
+    ('de_idnr', "German Personal Tax Number"),
+    ('de_stnr', "German Tax Number"),
+    ('de_vat', "German VAT Number"),
+    ('dk_cpr', "Danish Citizen Number"),
+    ('dk_cvr', "Danish VAT Number"),
+    ('do_cedula', "Dominican Republic National Identification Number"),
+    ('do_rnc', "Dominican Republic Tax"),
+    ('ec_ci', "Ecuadorian Personal Identity Code"),
+    ('ec_ruc', "Ecuadorian Tax Identification"),
+    ('ee_ik', "Estonian Personal ID Number"),
+    ('ee_kmkr', "Estonian VAT Number"),
+    ('ee_registrikood', "Estonian Organisation Registration Code"),
+    ('es_cif', "Spanish Company Tax"),
+    ('es_dni', "Spanish Personal Identity Codes"),
+    ('es_nie', "Spanish Foreigner Number"),
+    ('es_nif', "Spanish VAT Number"),
+    ('eu_at_02', "SEPA Identifier of the Creditor (AT-02)"),
+    ('eu_vat', "European VAT Number"),
+    ('fi_alv', "Finnish VAT Number"),
+    ('fi_associationid', "Finnish Association Identifier"),
+    ('fi_hetu', "Finnish Personal Identity Code"),
+    ('fi_veronumero', "Finnish Individual Tax Number"),
+    ('fi_ytunnus', "Finnish Business Identifier"),
+    ('fr_nif', "French Tax Identification Number"),
+    ('fr_nir', "French Personal Identification Number"),
+    ('fr_siren', "French Company Identification Number"),
+    ('fr_siret', "French Company Establishment Identification Number"),
+    ('fr_tva', "French VAT Number"),
+    ('gb_nhs',
+        "United Kingdom National Health Service Patient Identifier"),
+    ('gb_upn', "English Unique Pupil Number"),
+    ('gb_vat', "United Kingdom (and Isle of Man) VAT Number"),
+    ('gr_amka', "Greek Social Security Number"),
+    ('gr_vat', "Greek VAT Number"),
+    ('gt_nit', "Guatemala Tax Number"),
+    ('hr_oib', "Croatian Identification Number"),
+    ('hu_anum', "Hungarian VAT Number"),
+    ('id_npwp', "Indonesian VAT Number"),
+    ('ie_pps', "Irish Personal Number"),
+    ('ie_vat', "Irish VAT Number"),
+    ('il_hp', "Israeli Company Number"),
+    ('il_idnr', "Israeli Identity Number"),
+    ('in_aadhaar', "Indian Digital Resident Personal Identity Number"),
+    ('in_pan', "Indian Income Tax Identifier"),
+    ('is_kennitala',
+        "Icelandic Personal and Organisation Identity Code"),
+    ('is_vsk', "Icelandic VAT Number"),
+    ('it_codicefiscale', "Italian Tax Code for Individuals"),
+    ('it_iva', "Italian VAT Number"),
+    ('jp_cn', "Japanese Corporate Number"),
+    ('kr_brn', "South Korea Business Registration Number"),
+    ('kr_krn', "South Korean Resident Registration Number"),
+    ('lt_asmens', "Lithuanian Personal Number"),
+    ('lt_pvm', "Lithuanian VAT Number"),
+    ('lu_tva', "Luxembourgian VAT Number"),
+    ('lv_pvn', "Latvian VAT Number"),
+    ('mc_tva', "Monacan VAT Number"),
+    ('md_idno', "Moldavian Company Identification Number"),
+    ('mt_vat', "Maltese VAT Number"),
+    ('mu_nid', "Mauritian National Identifier"),
+    ('mx_rfc', "Mexican Tax Number"),
+    ('my_nric',
+        "Malaysian National Registration Identity Card Number"),
+    ('nl_brin', "Dutch School Identification Number"),
+    ('nl_bsn', "Dutch Citizen Identification Number"),
+    ('nl_btw', "Dutch VAT Number"),
+    ('nl_onderwijsnummer', "Dutch Student Identification Number"),
+    ('no_fodselsnummer',
+        "Norwegian Birth Number, the National Identity Number"),
+    ('no_mva', "Norwegian VAT Number"),
+    ('no_orgnr', "Norwegian Organisation Number"),
+    ('nz_ird', "New Zealand Inland Revenue Department Number"),
+    ('pe_cui', "Peruvian Identity Number"),
+    ('pe_ruc', "Peruvian Company Tax Number"),
+    ('pl_nip', "Polish VAT Number"),
+    ('pl_pesel', "Polish National Identification Number"),
+    ('pl_regon', "Polish Register of Economic Units"),
+    ('pt_nif', "Portuguese VAT Number"),
+    ('py_ruc', "Paraguay Tax Number"),
+    ('ro_cf', "Romanian VAT Number"),
+    ('ro_cnp', "Romanian Numerical Personal Code"),
+    ('ro_onrc', "Romanian ONRC Number"),
+    ('rs_pib', "Serbian Tax Identification"),
+    ('ru_inn', "Russian Tax identifier"),
+    ('se_orgnr', "Swedish Company Number"),
+    ('se_personnummer', "Swedish Personal Number"),
+    ('se_vat', "Swedish VAT Number"),
+    ('si_ddv', "Slovenian VAT Number"),
+    ('sk_dph', "Slovak VAT Number"),
+    ('sk_rc', "Slovak Birth Number"),
+    ('sm_coe', "San Marino National Tax Number"),
+    ('tr_tckimlik', "Turkish Personal Identification Number"),
+    ('ua_edrpou', "Ukrainian Identifier for Enterprises and Organizations"),
+    ('ua_rntrc', "Ukrainian Individual Taxpayer Registration Number"),
+    ('us_atin', "U.S. Adoption Taxpayer Identification Number"),
+    ('us_ein', "U.S. Employer Identification Number"),
+    ('us_itin', "U.S. Individual Taxpayer Identification Number"),
+    ('us_ptin', "U.S. Preparer Tax Identification Number"),
+    ('us_ssn', "U.S. Social Security Number"),
+    ('us_tin', "U.S. Taxpayer Identification Number"),
+    ('uy_ruc', "Uruguay Tax Number"),
+    ('ve_rif', "Venezuelan VAT Number"),
+    ('vn_mst', "Vietnam Tax Number"),
+    ('za_idnr', "South African Identity Document Number"),
+    ('za_tin', "South African Tax Identification Number"),
+    ]
+
+TAX_IDENTIFIER_TYPES = [
+    'ad_nrt',
+    'al_nipt',
+    'ar_cuit',
+    'at_uid',
+    'au_abn',
+    'au_acn',
+    'be_vat',
+    'bg_vat',
+    'by_unp',
+    'ch_vat',
+    'cl_rut',
+    'cn_uscc',
+    'co_rut',
+    'cr_cpj',
+    'cz_dic',
+    'de_vat',
+    'dk_cvr',
+    'do_rnc',
+    'ec_ruc',
+    'ee_kmkr',
+    'es_nif',
+    'eu_vat',
+    'fi_alv',
+    'fr_tva',
+    'gb_vat',
+    'gr_vat',
+    'gt_nit',
+    'hu_anum',
+    'id_npwp',
+    'ie_vat',
+    'il_hp',
+    'is_vsk',
+    'it_iva',
+    'jp_cn',
+    'kr_brn',
+    'lt_pvm',
+    'lu_tva',
+    'lv_pvn',
+    'mc_tva',
+    'md_idno',
+    'mt_vat',
+    'mx_rfc',
+    'nl_btw',
+    'no_mva',
+    'nz_ird',
+    'pe_ruc',
+    'pl_nip',
+    'pt_nif',
+    'py_ruc',
+    'ro_cf',
+    'rs_pib',
+    'ru_inn',
+    'se_vat',
+    'si_ddv',
+    'sk_dph',
+    'sm_coe',
+    'ua_edrpou',
+    'ua_rntrc',
+    'us_atin',
+    'us_ein',
+    'us_itin',
+    'us_ptin',
+    'us_ssn',
+    'us_tin',
+    'uy_ruc',
+    've_rif',
+    'vn_mst',
+    'za_tin',
+    ]
+
+
+class Identifier(sequence_ordered(), DeactivableMixin, ModelSQL, ModelView):
     'Party Identifier'
     __name__ = 'party.identifier'
     _rec_name = 'code'
     party = fields.Many2One('party.party', 'Party', ondelete='CASCADE',
         required=True, select=True,
         help="The party identified by this record.")
-    type = fields.Selection([
-            (None, ''),
-            ('ad_nrt', "Andorra Tax Number"),
-            ('al_nipt', "Albanian VAT Number"),
-            ('ar_cuit', "Argentinian Tax Number"),
-            ('at_businessid', "Austrian Company Register"),
-            ('at_tin', "Austrian Tax Identification"),
-            ('au_abn', "Australian Business Number"),
-            ('au_acn', "Australian Company Number"),
-            ('au_tfn', "Australian Tax File Number"),
-            ('be_vat', "Belgian Enterprise Number"),
-            ('bg_egn', "Bulgarian Personal Identity Codes"),
-            ('bg_pnf', "Bulgarian Number of a Foreigner"),
-            ('bg_vat', "Bulgarian VAT Number"),
-            ('br_cnpj', "Brazillian Company Identifier"),
-            ('br_cpf', "Brazillian National Identifier"),
-            ('ca_bn', "Canadian Business Number"),
-            ('ca_sin', "Canadian Social Insurance Number"),
-            ('ch_ssn', "Swiss Social Security Number"),
-            ('ch_uid', "Swiss Business Identifier"),
-            ('ch_vat', "Swiss VAT Number"),
-            ('cl_rut', "Chilean National Tax Number"),
-            ('cn_rit', "Chinese Resident Identity Card Number"),
-            ('co_nit', "Colombian Identity Code"),
-            ('co_rut', "Colombian Business Tax Number"),
-            ('cr_cpf', "Costa Rica Physical Person ID Number"),
-            ('cr_cr', "Costa Rica Foreigners ID Number"),
-            ('cu_ni', "Cuban Identity Card Number"),
-            ('cy_vat', "Cypriot VAT Number"),
-            ('cz_dic', "Czech VAT Number"),
-            ('cz_rc', "Czech National Identifier"),
-            ('de_handelsregisternummer', "German Company Register Number"),
-            ('de_idnr', "German Personal Tax Number"),
-            ('de_stnr', "German Tax Number"),
-            ('de_vat', "German VAT Number"),
-            ('dk_cpr', "Danish Citizen Number"),
-            ('dk_cvr', "Danish VAT Number"),
-            ('do_cedula', "Dominican Republic National Identification Number"),
-            ('do_rnc', "Dominican Republic Tax"),
-            ('ec_ci', "Ecuadorian Personal Identity Code"),
-            ('ec_ruc', "Ecuadorian Tax Identification"),
-            ('ee_ik', "Estonian Personcal ID number"),
-            ('ee_kmkr', "Estonian VAT Number"),
-            ('ee_registrikood', "Estonian Organisation Registration Code"),
-            ('es_cif', "Spanish Company Tax"),
-            ('es_dni', "Spanish Personal Identity Codes"),
-            ('es_nie', "Spanish Foreigner Number"),
-            ('es_nif', "Spanish VAT Number"),
-            ('eu_at_02', "SEPA Identifier of the Creditor (AT-02)"),
-            ('eu_vat', "European VAT Number"),
-            ('fi_alv', "Finnish VAT Number"),
-            ('fi_associationid', "Finnish Association Identifier"),
-            ('fi_hetu', "Finnish Personal Identity Code"),
-            ('fi_veronumero', "Finnish individual tax number"),
-            ('fi_ytunnus', "Finnish Business Identifier"),
-            ('fr_nif', "French Tax Identification Number"),
-            ('fr_nir', "French Personal Identification Number"),
-            # TODO: remove from party_siren
-            # ('fr_siren', "French Company Identification Number"),
-            ('fr_tva', "French VAT Number"),
-            ('gb_nhs',
-                "United Kingdom National Health Service Patient Identifier"),
-            ('gb_upn', "English Unique Pupil Number"),
-            ('gb_vat', "United Kingdom (and Isle of Man) VAT Number"),
-            ('gr_vat', "Greek VAT Number"),
-            ('gt_nit', "Guatemala Tax Number"),
-            ('hr_oib', "Croatian Identification Number"),
-            ('hu_anum', "Hungarian VAT Number"),
-            ('ie_pps', "Irish Personal Number"),
-            ('ie_vat', "Irish VAT Number"),
-            ('in_aadhaar', "Indian Digital Resident Personal Identity Number"),
-            ('in_pan', "Indian Income Tax Identifier"),
-            ('is_kennitala',
-                "Icelandic Personal and Organisation Identity Code"),
-            ('is_vsk', "Icelandic VAT Number"),
-            ('it_codicefiscale', "Italian Tax Code for Individuals"),
-            ('it_iva', "Italian VAT Number"),
-            ('jp_cn', "Japanese Corporate Number"),
-            ('lt_pvm', "Lithuanian VAT Number"),
-            ('lu_tva', "Luxembourgian VAT Number"),
-            ('lv_pvn', "Latvian VAT Number"),
-            ('mc_tva', "Monacan VAT Number"),
-            ('md_idno', "Moldavian Company Identification Number"),
-            ('mt_vat', "Maltese VAT Number"),
-            ('mu_nid', "Mauritian National Identifier"),
-            ('mx_rfc', "Mexican Tax Number"),
-            ('my_nric',
-                "Malaysian National Registration Identity Card Number"),
-            ('nl_brin', "Dutch School Identification Number"),
-            ('nl_bsn', "Dutch Citizen Identification Number"),
-            ('nl_btw', "Dutch VAT Number"),
-            ('nl_onderwijsnummer', "Dutch student identification number"),
-            ('no_mva', "Norwegian VAT Number"),
-            ('no_orgnr', "Norwegian Organisation Number"),
-            ('nz_ird', "New Zealand Inland Revenue Department Number"),
-            ('pe_cui', "Peruvian Identity Number"),
-            ('pe_ruc', "Peruvian Company Tax Number"),
-            ('pl_nip', "Polish VAT Number"),
-            ('pl_pesel', "Polish National Identification Number"),
-            ('pl_regon', "Polish Register of Economic Units"),
-            ('pt_nif', "Portuguese VAT Number"),
-            ('py_ruc', "Paraguay Tax Number"),
-            ('ro_cf', "Romanian VAT Number"),
-            ('ro_cnp', "Romanian Numerical Personal Code"),
-            ('rs_pib', "Serbian Tax Identification"),
-            ('ru_inn', "Russian Tax identifier"),
-            ('se_orgnr', "Swedish Company Number"),
-            ('se_vat', "Swedish VAT Number"),
-            ('si_ddv', "Slovenian VAT Number"),
-            ('sk_dph', "Slovak VAT Number"),
-            ('sk_rc', "Slovak Birth Number"),
-            ('sm_coe', "San Marino National Tax Number"),
-            ('tr_tckimlik', "Turkish Personal Identification Number"),
-            ('us_atin', "U.S. Adoption Taxpayer Identification Number"),
-            ('us_ein', "U.S. Employer Identification Number"),
-            ('us_itin', "U.S. Individual Taxpayer Identification Number"),
-            ('us_ptin', "U.S. Preparer Tax Identification Number"),
-            ('us_ssn', "U.S. Social Security Number"),
-            ('us_tin', "U.S. Taxpayer Identification Number"),
-            ('uy_ruc', "Uruguay Tax Number"),
-            ('ve_rif', "Venezuelan VAT Number"),
-            ('za_idnr', "South African Identity Document Number"),
-            ('za_tin', "South African Tax Identification Number"),
-            ], 'Type')
+    address = fields.Many2One(
+        'party.address', "Address", ondelete='CASCADE', select=True,
+        states={
+            'required': Eval('type_address', False),
+            'invisible': ~Eval('type_address', True),
+            },
+        domain=[
+            ('party', '=', Eval('party', -1)),
+            ],
+        help="The address identified by this record.")
+    type = fields.Selection('get_types', 'Type')
     type_string = type.translated('type')
+    type_address = fields.Function(
+        fields.Boolean("Type of Address"), 'on_change_with_type_address')
     code = fields.Char('Code', required=True)
 
     @classmethod
@@ -447,6 +638,7 @@ class Identifier(sequence_ordered(), ModelSQL, ModelView):
         Party = pool.get('party.party')
         cursor = Transaction().connection.cursor()
         party = Party.__table__()
+        table = cls.__table__()
 
         super().__register__(module_name)
 
@@ -458,7 +650,7 @@ class Identifier(sequence_ordered(), ModelSQL, ModelView):
                     party.id, party.vat_number, party.vat_country,
                     where=(party.vat_number != Null)
                     | (party.vat_country != Null)))
-            for party_id, number, country in cursor.fetchall():
+            for party_id, number, country in cursor:
                 code = (country or '') + (number or '')
                 if not code:
                     continue
@@ -474,6 +666,25 @@ class Identifier(sequence_ordered(), ModelSQL, ModelView):
             party_h.drop_column('vat_number')
             party_h.drop_column('vat_country')
 
+        # Migration from 5.8: Rename cn_rit into cn_ric
+        cursor.execute(*table.update([table.type], ['cn_ric'],
+                where=(table.type == 'cn_rit')))
+
+    @classmethod
+    def get_types(cls):
+        pool = Pool()
+        Configuration = pool.get('party.configuration')
+        configuration = Configuration(1)
+        return [(None, '')] + configuration.get_identifier_types()
+
+    @classmethod
+    def _type_addresses(cls):
+        return {'fr_siret'}
+
+    @fields.depends('type')
+    def on_change_with_type_address(self, name=None):
+        return self.type in self._type_addresses()
+
     @fields.depends('type', 'code')
     def on_change_with_code(self):
         if self.type and '_' in self.type:
@@ -608,6 +819,7 @@ class Replace(Wizard):
         pool = Pool()
         Address = pool.get('party.address')
         ContactMechanism = pool.get('party.contact_mechanism')
+        Identifier = pool.get('party.identifier')
         transaction = Transaction()
 
         self.check_similarity()
@@ -620,6 +832,9 @@ class Replace(Wizard):
         ContactMechanism.write(list(source.contact_mechanisms), {
                 'active': False,
                 })
+        Identifier.write(list(source.identifiers), {
+                'active': False,
+                })
         source.replaced_by = destination
         source.active = False
         source.save()
@@ -627,9 +842,16 @@ class Replace(Wizard):
         cursor = transaction.connection.cursor()
         for model_name, field_name in self.fields_to_replace():
             Model = pool.get(model_name)
+            field = getattr(Model, field_name)
             table = Model.__table__()
             column = Column(table, field_name)
-            where = column == source.id
+            if field._type == 'reference':
+                source_value = str(source)
+                destination_value = str(destination)
+            else:
+                source_value = source.id
+                destination_value = destination.id
+            where = column == source_value
 
             if transaction.database.has_returning():
                 returning = [table.id]
@@ -640,7 +862,7 @@ class Replace(Wizard):
 
             cursor.execute(*table.update(
                     [column],
-                    [destination.id],
+                    [destination_value],
                     where=where,
                     returning=returning))
 
@@ -655,6 +877,7 @@ class Replace(Wizard):
         return [
             ('party.address', 'party'),
             ('party.contact_mechanism', 'party'),
+            ('party.identifier', 'party'),
             ]
 
 
@@ -667,7 +890,6 @@ class ReplaceAsk(ModelView):
         domain=[
             ('id', '!=', Eval('source', -1)),
             ],
-        depends=['source'],
         help="The party that replaces.")
 
     @classmethod
@@ -780,7 +1002,8 @@ class Erase(Wizard):
                 ['type', 'code'],
                 [None, '****']),
             (Address, [('party', '=', party_id)], True,
-                ['name', 'street', 'zip', 'city', 'country', 'subdivision'],
+                ['name', 'street', 'postal_code', 'city',
+                    'country', 'subdivision'],
                 [None, None, None, None, None, None]),
             (ContactMechanism, [('party', '=', party_id)], True,
                 ['value', 'name', 'comment'],
@@ -792,7 +1015,8 @@ class Erase(Wizard):
         pool = Pool()
         Attachment = pool.get('ir.attachment')
         Note = pool.get('ir.note')
-        return [Attachment, Note]
+        Avatar = pool.get('ir.avatar')
+        return [Attachment, Note, Avatar]
 
 
 class EraseAsk(ModelView):
